/*
 *   SIMPLL - A tool for PLL syntesizer loop filter optimization 
 *
 *      (c) by Lapo Pieri IK5NAX  2006
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send bugs reports, comments, critique, etc, to (ik5nax #AT# amsat.org) or
 *  (lapo_pieri #AT# virgilio.it)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "main.h"

double Icp, Kvco, frif;                                     /* Parametri pll */
int N;
double fstart, fstop;                                   /* Parametri analisi */
char lfcn[256];                          /* Nome file .ckt (.cir) loopfilter */

void double2prefs(double, int, char *, int);

int main(int argc, char **argv){
  int l=1, lc=0, rv, head, oc, isgnuplot, invertlf, isacc, atn=0, clo,
    savetmp=0, i;
  char linebuff[256], *lb, cfn[16], dfn[16], cgpn[16], buff[256];
  FILE *in, *lf, *cf, *df, *cgp, *acc;
  double freq, mod, ph, oldfreq, oldmod, oldph, xi, xs, yi, ys, pm, fpm, a;

  if(argc<2){
    printf("Usage:\n\
spc <filename.dat> [<accfile>] # for closed loop analysis\n\
spo <filename.dat> [<accfile>] # for open loop analysis\n");
    return 0;
  }

  while(1){
    clo=getopt(argc, argv, "s");
    if(clo==-1) break;
    switch(clo){
    case 's': savetmp=1; break; 
    }
  }

  if(strcmp(argv[optind+0], "-")==0)
    in=stdin;
  else{
    if((in=(FILE *)fopen(argv[optind+0], "r"))==NULL){
      printf("Error opening input file: %s\n", argv[optind+0]);
      return 1;
    }
  }

  if(argc-optind>1){
    isacc=1;
    if((acc=(FILE *)fopen(argv[optind+1], "a+"))==NULL){
      printf("Error opening accumulation file: %s\n", argv[optind+1]);
      return 1;
    } 
    while(1){
      if(fgets(linebuff, 255, acc)==NULL)
	break;
      if((lb=strstr(linebuff, __atm))!=NULL){
	lb+=strlen(__atm);
	sscanf(lb, "%d", &atn);
      }
    }
   
  }
  else
    isacc=0;

  /* Rilevazione presenza ngspice */
  rv=system("ngspice -b 2>/dev/null >/dev/null");
  switch(rv){
  case 0x7f00: printf("ngspice not present!\n"); return 1;
  case 0x100:
    break;
  default: 
    printf("Strange ngspice return value (0x%x), trying to proceed...\n", rv);
  }

  /* Rilevazione presenza GNUPLOT */
  rv=system("gnuplot --version 2>/dev/null >/dev/null");
  switch(rv){
  case 0x7f00:
    printf("GNUPLOT not present!\n"); isgnuplot=0;
    break;
  case 0:
  case 0x0100:
    isgnuplot=1;
    break;
  default:
    printf("Strange GNUPLOT return value (0x%x), trying to proceed...\n", rv);
    isgnuplot=1;
  }


  /* Impostazione analisi anello aperto o chiuso */
  lb=strrchr(argv[0], '/'); 
  if(lb==NULL) 
    lb=argv[0];
  else
    lb++;
  if(strncmp(lb, "spo", 3)==0)
    oc=1;                                           /* analisi anello aperto */
  else if(strncmp(lb, "spc", 3)==0)
    oc=-1;                                          /* analisi anello chiuso */
  else{
    printf("Unable to understand if analysis has to be performed open or \
closed loop\n");
    return 1;
  }


  /* Scansione file dati */
  while(l){
    if(fgets(linebuff, 255, in)==NULL)
      break;
    lc++;
    lb=linebuff;
    while(*lb==' ' || *lb=='\t') lb++;
    if(strncasecmp(lb, "Icp", sizeof("Icp")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      sscanf(lb, "%lf", &Icp);
    }
    else if(strncasecmp(lb, "Kvco", sizeof("Kvco")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      sscanf(lb, "%lf", &Kvco);
    }
    else if(strncasecmp(lb, "N", sizeof("N")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      sscanf(lb, "%d", &N);
    }
    else if(strncasecmp(lb, "frif", sizeof("frif")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      sscanf(lb, "%lf", &frif);
    }
    else if(strncasecmp(lb, "loopflt", sizeof("loopflt")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      strncpy(lfcn, lb, sizeof(lfcn)-1);
    }
    else if(strncasecmp(lb, "invlf", sizeof("invfl")-1)==0){
      while(*lb!=' ' && *lb!='\t') lb++;
      while(*lb==' ' || *lb=='\t') lb++;
      if(sscanf(lb, "%d", &invertlf)==0){
	printf("Wrong value for invlf: %s\n", linebuff);
	return 1;
      }
      if(invertlf!=0 && invertlf!=1){
	printf("Wrong value for invlf: %s\n", linebuff);
	printf("Allowed value are 0 for non inverting and 1 for inverting \
loop filter\n");
	return 1;
      }	

    }
    else{
      printf("Error in parameter file on line %d:\n%s\n", lc, linebuff);
      return 1;
    }
  }


  /* Apertura file circuito filtro anello */
  lb=lfcn;
  while(*lb!='\0'){
    lb++;      
    if(*lb=='\n' || *lb=='\r'){
      *lb='\0';
    }
  }
  if((lf=(FILE *)fopen(lfcn, "r"))==NULL){
    printf("Error opening loop filter file: %s\n", lfcn);
    return 1;
  }

  /* Apertura file (temporaneo) circuito complessivo */
  strcpy(cfn, "spCXXXXXX"); mkstemp(cfn);
  if((cf=(FILE *)fopen(cfn, "w"))==NULL){
    printf("Error opening loop filter file: %s\n", cfn);
    return 1;
  }

  /* Scrittura file circuito complessivo */
  fprintf(cf, "%s\n", oc==1?__comment_o:__comment_c);


  /* Copia file lfcn in cfn */
  while(1){
    if(fgets(linebuff, 255, lf)==NULL)
      break;
    fputs(linebuff, cf);
  }


  /* DAFARE: ricavare dal file dati ? */
  fstart=1; fstop=1e6;

  /* Scrittura parte comune circuito di simulazione */
  fprintf(cf, oc==1?bco:bcc, Icp, Kvco, 1./(double)N, fstart, fstop);
                          /* DAFARE: aumentare la parametrizzazione */

  /* Creazione nome file temporaneo per dati in uscita dall'ngspice */
  strcpy(dfn, "spDXXXXXX"); mkstemp(dfn);


  /* Esecuzione ngspice */
  snprintf(buff, 255, "ngspice -b %s >%s", cfn, dfn);
  fclose(cf); system(buff); 


  /* Apertura file dati in uscita dall'ngspice */
  if((df=(FILE *)fopen(dfn, "r"))==NULL){
    printf("Error analysis output file: %s\n", dfn);
    return 1;
  }


  /* Analisi dei risultai e post processazione */
  if(oc==1){                                                /* anello aperto */
    /* Scansione file df ricerca passaggio per lo zero del modulo*/
    head=1; lc=0; l=1;
    oldmod=1e6;
    while(l==1){
      if(fgets(linebuff, 255, df)==NULL)
	l=-1;
      lc++;
      lb=linebuff;
      while(*lb==' ' || *lb=='\t') lb++;
      i=0;
      while(lb[i]!='\0'){
	if(lb[i]==',')
	  lb[i]=' ';
	i++;
      }
      
      rv=sscanf(lb, "%*d %lf %*f %lf %lf", &freq, &mod, &ph);
      if(rv==0 || rv==EOF){
	continue;
      }
      if(oldmod>0. && mod<=0.){
	l=0;
      }
      else{
	oldfreq=freq; oldmod=mod; oldph=ph;
      }
    }
    
    if(l==-1){
      printf("Unable to find mod=0dB point, check output file %s\n", dfn);
      return 1;
    }
    
    /*
      Interpolazione lineare fra i due punti del passaggio per lo zero del
      guadagno per trovare la frequenza di passaggio per lo zero e i margine
      di fase
    */
    ph=ph/M_PI*180.; oldph=oldph/M_PI*180.;
    xi=oldfreq; xs=freq;
    yi=10*pow(oldmod/20, 10); ys=10*pow(mod/20, 10);
    fpm=xi*(ys*xi-yi*xs)/(ys*xi-yi*xs-yi);
    a=(fpm-xi)/(xs-xi);
    pm=(1.-a)*oldph+a*ph+(invertlf==1?0.:180.);  /* 
						    Attenzione: il margine di
						    fase e` 180 piu`
						    del valore della fase a
						    ampiezza 0dB; se pero`
						    il filtro e` invertente
						    si deve negare ancora
						 */
    
    
    /* (Ri)scansione file df ricerca punto a frif */
    rewind(df); head=1; lc=0; l=1; oldfreq=fstart;
    while(l==1){
      if(fgets(linebuff, 255, df)==NULL)
	l=-1;
      lc++;
      lb=linebuff;
      while(*lb==' ' || *lb=='\t') lb++;
      i=0;
      while(lb[i]!='\0'){
	if(lb[i]==',')
	  lb[i]=' ';
	i++;
      }
      
      rv=sscanf(lb, "%*d %lf %*f %lf %lf", &freq, &mod, &ph);
      if(rv==0 || rv==EOF){
	continue;
      }

      if(oldfreq<frif && freq>=frif){
	l=0;
      }
      else{
	oldfreq=freq; oldmod=mod; oldph=ph;
      }
    }
    fclose(df);
    
    if(l==-1){
      printf("Unable to find frif point\n");
    }  
    
    /* Accumulo circuiti (per ora solo nell'analisi a anello aperto) */
    if(isacc){
      atn++;
      fprintf(acc, "%s%d\n", __atm, atn);
      rewind(lf);
      while(1){
	if(fgets(linebuff, 255, lf)==NULL)
	  break;
	fputs(linebuff, acc);
      }
    }
    fclose(lf);
    
    /* Stampa risultati (e accumulo)*/
    double2prefs(pm, 1, linebuff, 0);
    printf("Phase margin= %s @ ", linebuff);
    if(isacc)
      fprintf(acc, "\t\t\t\t%sdeg @ ", linebuff);
    double2prefs(fpm, 2, linebuff, 0);
    printf("%sHz\t", linebuff);
    if(isacc)
      fprintf(acc, "%sHz\t", linebuff);
    if(l!=-1){
      a=(frif-oldfreq)/(freq-oldfreq);
      double2prefs((1.-a)*oldmod+a*mod, 1, linebuff, 0);
      printf("Atten[@frif]= %sdB\n", linebuff);
    if(isacc)
      fprintf(acc, "%sdB\n", linebuff);
    }
    else{
      printf("\n");
    if(isacc)
      fprintf(acc, "\n");
    }

    if(isacc){
      printf("Comment? (max 255 characters)");
      fgets(linebuff, 255, stdin);
      fprintf(acc, "######## %s\n", linebuff);
    }

  }
  else if(oc==-1){

    if(isgnuplot){
      /* Apertura file (temporaneo) comandi gnuplot */
      strcpy(cgpn, "cgpXXXXXX"); mkstemp(cgpn);
      if((cgp=(FILE *)fopen(cgpn, "w"))==NULL){
	printf("Error opening gnuplot command temporary file: %s\n", cgpn);
	return 1;
      }
      fprintf(cgp, cmdgpt, dfn); fclose(cgp);
      
      sprintf(linebuff, "gnuplot -persist %s", cgpn);
      system(linebuff);
      remove(cgpn);
    }
    else{
      sprintf(linebuff, "cat %s", dfn);
      system(linebuff);
    }

  }

  /* Chiusura file e scancellazione file temporanei*/
  if(savetmp==0){
    remove(cfn); remove(dfn);
  }
  if(isacc) fclose(acc);

  return 0;
}

